#ifndef _REV_HASH_TABLE_H
#define _REV_HASH_TABLE_H

#include <math.h>
#include <vector>
#include <list>
#include "globals.h"
#include "util.h"

using namespace std;

#define MAX_KEY       255
#define NUM_DIVISIONS (sizeof(KeyType))

class RevHashTable {

 private:
  ValueType                     *chgLst; // made a class member to avoid repeating memory alloc
  
 public:
  uint32                         lookupTable[8][256]; // IP key sections -> hash index
  ValueType                     *valueTable;  // actual data goes here
  vector<vector<uint8> >         revLookupTable[8];

  uint32 numBins;             // 1<<inxBits
  uint32 binMask;             // numBins-1
  uint32 bitsPerDiv;          // inxBits / NUM_DIVISIONS
  uint32 binsPerDiv;          // 1<<bitsPerDiv
  uint32 divMask;             // binsPerDiv-1

  // Helper Functions
  void PopulateLookupTable(int div, int seed);
  
 public:
  // Constructor
  RevHashTable(int seed, int inxBits);  

  // Destructor
  ~RevHashTable(); 

  // Accessor
  uint32 GetNumBins() const { return numBins; }
  
  // Member Functions
  void AddValue(KeyType key, ValueType val)
    {
      valueTable[GetIndex(key)] += val;
    }
  
  void RemoveValue(KeyType key, ValueType val)
    {
      valueTable[GetIndex(key)] -= val;
    }
  
  ValueType GetValue(KeyType key)
    {
      return valueTable[GetIndex(key)];
    }

  IndexType GetIndex(uint32 key)
    {
      // key is in permuted format
      // XXX: assume values in lookupTable[][] has been shifted properly
      int32views h;
      h.as_int32 = key;
#ifdef BIGENDIAN
      return
        lookupTable[0][h.as_int8s[3]] |
        lookupTable[1][h.as_int8s[2]] |
        lookupTable[2][h.as_int8s[1]] |
        lookupTable[3][h.as_int8s[0]];
#else
      return
        lookupTable[0][h.as_int8s[0]] |
        lookupTable[1][h.as_int8s[1]] |
        lookupTable[2][h.as_int8s[2]] |
        lookupTable[3][h.as_int8s[3]];
#endif
    }
  
  IndexType GetIndex(uint64 key)
    {
      // key is in permuted format
      // XXX: assume values in lookupTable[][] has been shifted properly
      int64views h;
      h.as_int64 = key;
#ifdef BIGENDIAN
      return
        lookupTable[0][h.as_int8s[7]] |
        lookupTable[1][h.as_int8s[6]] |
        lookupTable[2][h.as_int8s[5]] |
        lookupTable[3][h.as_int8s[4]] |
        lookupTable[4][h.as_int8s[3]] |
        lookupTable[5][h.as_int8s[2]] |
        lookupTable[6][h.as_int8s[1]] |
        lookupTable[7][h.as_int8s[0]];
#else
      return
        lookupTable[0][h.as_int8s[0]] |
        lookupTable[1][h.as_int8s[1]] |
        lookupTable[2][h.as_int8s[2]] |
        lookupTable[3][h.as_int8s[3]] |
        lookupTable[4][h.as_int8s[4]] |
        lookupTable[5][h.as_int8s[5]] |
        lookupTable[6][h.as_int8s[6]] |
        lookupTable[7][h.as_int8s[7]];
#endif
    }
  
  void GetPossibleKeys(list<uint8> &keys, int div, IndexType bin);

  // Change detection compares the absolute value of the difference of each
  // of the bins to the threshold. It returns a sorted vector containing
  // the indices of the bins exceeding the threshold.
  void ChangeDetection(list<IndexType> &changes, const RevHashTable *otherTbl, 
		       double threshold);

  double FindThreshold(const RevHashTable *otherTbl,
		       int numAnoms, double minThreshold);
};

#endif
